use eframe::{
    egui::{Color32, Pos2, Ui},
    epaint::{CircleShape, PathShape, PathStroke},
};
use egui_plot::{Line, PlotPoint, PlotPoints, PlotUi};

use crate::{
    application::Application,
    geometry::{elements::circle::Circle, DrawableRaw},
};

use super::{draw_floating_area_on_canvas, draw_on_plot_canvas, CanvasColour, Drawer, PlotDrawer};

pub fn draw_geometries_(ui: &mut PlotUi, app: &mut Application) {
    for (file_name, geo) in &app.outlines {
        let selected = &app.selection == file_name;

        // draw outline path
        for path in geo.paths().iter() {
            draw_on_plot_canvas(
                ui,
                app,
                PlotDrawer::Closure(&|ui| {
                    // draw outline path
                    let mut points = path
                        .iter()
                        .map(|p| [p.x(), p.y()])
                        .collect::<Vec<[f64; 2]>>();

                    // calculate line width according to zoom factor
                    let transform = ui.transform();
                    let p1: Pos2 = transform.position_from_point(&PlotPoint::from(points[0]));
                    let p2: Pos2 = transform.position_from_point(&PlotPoint::from([
                        points[0][0] - geo.stroke as f64,
                        points[0][1],
                    ]));
                    let width = (p1.x - p2.x).abs();

                    points.push(points[0]);
                    points.push(points[1]);
                    let line = Line::new(PlotPoints::from(points))
                        .width(width)
                        .color(CanvasColour::Outline.to_colour32(selected));

                    ui.line(line)
                }),
            );
        }

        // draw point shapes
        for point in geo.points().iter() {
            draw_on_plot_canvas(
                ui,
                app,
                PlotDrawer::Closure(&|ui| {
                    let circle = Circle::new(*point, geo.stroke.into(), None);
                    circle.draw_egui_plot(ui, CanvasColour::Outline, selected);
                }),
            );
            //     draw_floating_area_on_canvas(
            //         ui,
            //         app,
            //         point,
            //         ("GeometryAreaPoint", name, i),
            //         Drawer::Closure(&|ui| {
            //             // draw point shape
            //             ui.painter().add(CircleShape::filled(
            //                 point.invert_y().into(),
            //                 geo.stroke,
            //                 if selected {
            //                     Color32::DARK_BLUE.gamma_multiply(0.5)
            //                 } else {
            //                     Color32::DARK_BLUE
            //                 },
            //             ));
            //         }),
            //     );
        }
    }

    // for circle in excellon.holes.iter() {
    //     draw_on_plot_canvas(
    //         ui,
    //         app,
    //         PlotDrawer::Drawable((
    //             &Element::Circle(circle.to_owned()),
    //             CanvasColour::Excellon,
    //             selected,
    //         )),
    //     );
    // }
}

pub fn draw_geometries(ui: &mut Ui, app: &mut Application) {
    for (name, geo) in &app.outlines {
        let selected = &app.selection == name;

        // draw outline path
        for (i, path) in geo.paths().iter().enumerate() {
            draw_floating_area_on_canvas(
                ui,
                app,
                {
                    let origin = path.iter().next().unwrap();
                    Pos2::new(origin.x() as f32, origin.y() as f32)
                },
                ("GeometryAreaOutline", name, i),
                Drawer::Closure(&|ui| {
                    // draw outline path
                    ui.painter().add(PathShape::closed_line(
                        path.iter()
                            .map(|v| Pos2::new(v.x() as f32, -v.y() as f32))
                            .collect::<Vec<Pos2>>(),
                        PathStroke::new(
                            geo.stroke,
                            if selected {
                                Color32::DARK_BLUE.gamma_multiply(0.5)
                            } else {
                                Color32::DARK_BLUE
                            },
                        ),
                    ));
                }),
            );
        }

        // draw point shapes
        for (i, point) in geo.points().iter().enumerate() {
            draw_floating_area_on_canvas(
                ui,
                app,
                point,
                ("GeometryAreaPoint", name, i),
                Drawer::Closure(&|ui| {
                    // draw point shape
                    ui.painter().add(CircleShape::filled(
                        point.invert_y().into(),
                        geo.stroke,
                        if selected {
                            Color32::DARK_BLUE.gamma_multiply(0.5)
                        } else {
                            Color32::DARK_BLUE
                        },
                    ));
                }),
            );
        }
    }
}
