use eframe::{
    egui::{Color32, Pos2, Ui},
    epaint::{PathShape, PathStroke},
};
use egui_plot::{Line, PlotPoints, PlotUi};

use crate::{application::Application, geometry::DrawableRaw};

use super::{draw_floating_area_on_canvas, draw_on_plot_canvas, CanvasColour, Drawer, PlotDrawer};

pub fn draw_gerbers_(ui: &mut PlotUi, app: &mut Application) {
    for (file_name, (_, geo)) in &app.gerbers {
        let selected = &app.selection == file_name;

        for geometry in geo.apertures.iter() {
            draw_on_plot_canvas(
                ui,
                app,
                PlotDrawer::Drawable((geometry, CanvasColour::Copper, selected)),
            );
        }

        for line in geo.paths.iter() {
            draw_on_plot_canvas(
                ui,
                app,
                PlotDrawer::Closure(&|ui| line.draw_egui_plot(ui, CanvasColour::Copper, selected)),
            );
        }

        // draw union path
        for path in geo.outline_union.iter() {
            let mut points = path
                .iter()
                .map(|p| [p.x(), p.y()])
                .collect::<Vec<[f64; 2]>>();

            points.push(points[0]);
            let line = Line::new(PlotPoints::from(points))
                .color(CanvasColour::CopperOutline.to_colour32(selected));

            ui.line(line)
        }
    }
}

pub fn draw_gerbers(ui: &mut Ui, app: &mut Application) {
    for (file_name, (name, geo)) in &app.gerbers {
        let selected = &app.selection == file_name;

        for (i, geometry) in geo.apertures.iter().enumerate() {
            draw_floating_area_on_canvas(
                ui,
                app,
                geometry.canvas_pos(),
                ("GerberArea", name, i),
                Drawer::Drawable((geometry, selected)),
            );
        }

        for (i, line) in geo.paths.iter().enumerate() {
            draw_floating_area_on_canvas(
                ui,
                app,
                line.canvas_pos(),
                ("LinePath", name, i),
                Drawer::Closure(&|ui| line.draw_egui(ui, selected)),
            );
        }

        // draw union path
        for (i, path) in geo.outline_union.iter().enumerate() {
            draw_floating_area_on_canvas(
                ui,
                app,
                {
                    let origin = path.iter().next().unwrap();
                    Pos2::new(origin.x() as f32, origin.y() as f32)
                },
                ("UnionOutline", name, i),
                Drawer::Closure(&|ui| {
                    ui.painter().add(PathShape::closed_line(
                        path.iter()
                            .map(|v| Pos2::new(v.x() as f32, -v.y() as f32))
                            .collect::<Vec<Pos2>>(),
                        PathStroke::new(
                            0.1_f32,
                            if selected {
                                Color32::LIGHT_RED
                            } else {
                                Color32::BLACK
                            },
                        ),
                    ));
                }),
            );
        }
    }
}
