pub mod excellons;
pub mod geometries;
pub mod gerbers;
mod live_position;

use std::hash::Hash;

use eframe::{
    egui::{self, Color32, Pos2, Rounding, Stroke, Ui},
    epaint::RectShape,
};
use egui_plot::PlotUi;
use excellons::{draw_excellons, draw_excellons_};
use geometries::{draw_geometries, draw_geometries_};
use gerbers::{draw_gerbers, draw_gerbers_};
use live_position::draw_live_position;

use crate::geometry::elements::Element;

use super::Application;

const COPPER_COLOR: Color32 = Color32::from_rgb(0xCA, 0xF6, 0x8F);
const COPPER_COLOR_SELECTED: Color32 = Color32::YELLOW;
const COPPER_OUTLINE: Color32 = Color32::BLACK;
const COPPER_OUTLINE_SELECTED: Color32 = Color32::LIGHT_RED;
const EXCELLON_COLOR: Color32 = Color32::BROWN;
const EXCELLON_COLOR_SELECTED: Color32 = Color32::from_rgba_premultiplied(65, 42, 42, 128);
const OUTLINE_COLOR: Color32 = Color32::DARK_BLUE;
const OUTLINE_COLOR_SELECTED: Color32 = Color32::from_rgba_premultiplied(0, 0, 139, 128);

pub enum CanvasColour {
    Copper,
    CopperOutline,
    Excellon,
    Outline,
}

impl CanvasColour {
    pub fn to_colour32(&self, selected: bool) -> Color32 {
        match (self, selected) {
            (CanvasColour::Copper, true) => COPPER_COLOR_SELECTED,
            (CanvasColour::Copper, false) => COPPER_COLOR,
            (CanvasColour::CopperOutline, true) => COPPER_OUTLINE_SELECTED,
            (CanvasColour::CopperOutline, false) => COPPER_OUTLINE,
            (CanvasColour::Excellon, true) => EXCELLON_COLOR_SELECTED,
            (CanvasColour::Excellon, false) => EXCELLON_COLOR,
            (CanvasColour::Outline, true) => OUTLINE_COLOR_SELECTED,
            (CanvasColour::Outline, false) => OUTLINE_COLOR,
        }
    }
}

pub fn draw_canvas(ui: &mut Ui, app: &mut Application) {
    egui_plot::Plot::new("ApplicationPlot")
        // .view_aspect(2.0)
        .data_aspect(1.0)
        .show(ui, |plot_ui| {
            draw_gerbers_(plot_ui, app);
            draw_excellons_(plot_ui, app);
            draw_geometries_(plot_ui, app);
        });

    // draw_live_position(ui, app);

    // ui.painter().add(RectShape::stroke(
    //     app.canvas.1,
    //     Rounding::same(0.),
    //     Stroke::new(0.2, Color32::BLACK),
    // ));

    // draw_gerbers(ui, app);
    // draw_excellons(ui, app);
    // draw_geometries(ui, app);
}

pub enum Drawer<'a> {
    Drawable((&'a Element, bool)),
    Closure(&'a dyn Fn(&mut Ui)),
}

fn draw_floating_area_on_canvas(
    ui: &mut Ui,
    app: &Application,
    pos: impl Into<Pos2>,
    name: impl Hash,
    drawer: Drawer,
) {
    let window_layer = ui.layer_id();
    let id = egui::Area::new(app.canvas.0.with(name))
        .current_pos(pos)
        // .order(egui::Order::Middle)
        .show(ui.ctx(), |ui| {
            ui.set_clip_rect(app.test_transform.inverse() * app.canvas.1);
            match drawer {
                Drawer::Drawable((t, selected)) => t.draw_egui(ui, selected),
                Drawer::Closure(fun) => fun(ui),
            }
        })
        .response
        .layer_id;

    ui.ctx().set_transform_layer(id, app.test_transform);
    ui.ctx().set_sublayer(window_layer, id);
}

pub enum PlotDrawer<'a> {
    Drawable((&'a Element, CanvasColour, bool)),
    Closure(&'a dyn Fn(&mut PlotUi)),
}

fn draw_on_plot_canvas(ui: &mut PlotUi, app: &Application, drawer: PlotDrawer) {
    match drawer {
        PlotDrawer::Drawable((t, colour, selected)) => t.draw_egui_plot(ui, colour, selected),
        PlotDrawer::Closure(fun) => fun(ui),
    }
}
