use std::f64::consts::PI;

use eframe::{
    egui::{Pos2, Stroke, Ui},
    epaint::{CircleShape, PathShape, PathStroke},
};
use egui_plot::{PlotPoints, Polygon};

use crate::{
    application::CanvasColour,
    geometry::{helpers::semi_circle, ClipperPath, ClipperPaths},
};

use super::super::{
    point::{convert_to_unit, Point},
    DrawableRaw, Unit,
};

#[derive(Debug, Clone)]
pub struct LinePath {
    pub points: Vec<Point>,
    diameter: f64,
    pub outline: ClipperPaths,
}

impl LinePath {
    pub fn new() -> Self {
        Self {
            points: Vec::new(),
            diameter: 0.,
            outline: ClipperPaths::new(vec![]),
        }
    }

    pub fn is_empty(&self) -> bool {
        self.points.is_empty()
    }

    pub fn add(&mut self, point: Point) {
        self.points.push(point);
    }

    pub fn set_stroke(&mut self, width: f64) {
        self.diameter = width
    }

    pub fn finalize(&mut self, stroke_width: f64) {
        self.diameter = stroke_width;
        self.outline = self.create_outline();
    }

    pub fn to_unit(&self, origin: Unit, to: Unit) -> Self {
        let mut converted = Self {
            points: self.points.iter().map(|p| p.to_unit(origin, to)).collect(),
            diameter: convert_to_unit(self.diameter, origin, to),
            outline: ClipperPaths::new(vec![]),
        };
        converted.outline = converted.create_outline();

        converted
    }

    fn create_outline(&self) -> ClipperPaths {
        let mut paths: Vec<ClipperPath> = Vec::new();

        for (index, point) in self.points.iter().enumerate() {
            if index > 0 {
                paths.push(
                    create_outline_between_points(self.points[index - 1], *point, self.diameter)
                        .into(),
                )
            }
        }

        ClipperPaths::new(paths)
    }
}

impl DrawableRaw for &LinePath {
    fn canvas_pos(&self) -> Point {
        if let Some(point) = self.points.first() {
            point.to_owned()
        } else {
            Point::new(0., 0.)
        }
    }

    fn draw_egui_plot(&self, ui: &mut egui_plot::PlotUi, colour: CanvasColour, selected: bool) {
        if let Some(path) = self.outline.get(0) {
            let mut points: Vec<[f64; 2]> = path.iter().map(|p| [p.x(), p.y()]).collect();
            points.push(points[0]);

            let poly = Polygon::new(PlotPoints::from(points))
                .fill_color(colour.to_colour32(selected))
                .stroke(Stroke::NONE);
            ui.polygon(poly);
        }
    }

    fn draw_egui(&self, ui: &mut Ui, selected: bool) {
        ui.painter().add(PathShape::line(
            self.points
                .iter()
                .map(|v| (*v).invert_y().into())
                .collect::<Vec<Pos2>>(),
            PathStroke::new(
                self.diameter as f32,
                CanvasColour::Copper.to_colour32(selected),
            ),
        ));

        for point in &self.points {
            ui.painter().add(CircleShape::filled(
                point.invert_y().into(),
                self.diameter as f32 / 2.,
                CanvasColour::Copper.to_colour32(selected),
            ));
        }
    }

    fn to_paths(&self) -> ClipperPaths {
        self.outline.clone()
    }

    fn outline(&self) -> ClipperPath {
        self.outline.first().unwrap().clone()
    }
}

fn create_outline_between_points(point1: Point, point2: Point, width: f64) -> Vec<(f64, f64)> {
    let line_vec = point2 - point1; // vector between start and end of line
    let normalized = line_vec.normalize();

    let angle =
        (normalized.x).acos() * (180. / PI) * if normalized.y < 0. { -1. } else { 1. } + 90.;

    let mut outline: Vec<(f64, f64)> = Vec::new();

    outline.append(&mut semi_circle(point1, width, angle));
    outline.append(&mut semi_circle(point2, width, angle + 180.));

    outline
}
