use eframe::{
    egui::{Rect, Rounding, Stroke, Ui, Vec2},
    epaint::RectShape,
};
use egui_plot::{PlotPoints, Polygon};

use crate::{
    application::CanvasColour,
    geometry::{ClipperPath, ClipperPaths},
};

use super::rectangle::Rectangle;
use super::{
    super::{
        helpers::{create_circular_path, semi_circle, CircleSegment},
        point::{convert_to_unit, Point},
        DrawableRaw, Unit,
    },
    circle::Circle,
};

#[derive(Debug)]
pub struct Obround {
    position: Point,
    x: f64,
    y: f64,
    rounding: f64,
    outline: ClipperPath,
    rectangle: Rectangle,
    hole_diameter: Option<f64>,
}

impl Obround {
    pub fn new(position: Point, x: f64, y: f64, hole_diameter: Option<f64>) -> Self {
        let diameter = if x < y { x } else { y };
        let outline = if x == y {
            create_circular_path(&position, x, CircleSegment::Full)
        } else {
            let mut path: Vec<(f64, f64)> = Vec::new();
            // check if obround is round to x or y
            if x < y {
                // round on y axis
                path.append(&mut semi_circle(position - Point::new(0., y / 4.), x, 180.));
                path.append(&mut semi_circle(position + Point::new(0., y / 4.), x, 0.));
            } else {
                // TODO round on x axis -> check for correctness!!!!!!!
                path.append(&mut semi_circle(position - Point::new(0., x / 4.), y, 270.));
                path.append(&mut semi_circle(position + Point::new(0., x / 4.), y, 90.));
            }

            path
        };

        Self {
            position,
            x: x,
            y: y,
            rounding: diameter,
            outline: outline.into(),
            rectangle: Rectangle::new(position, x, y, hole_diameter),
            hole_diameter: hole_diameter,
        }
    }

    pub fn from_aperture_obround(aperture: &gerber_types::Rectangular, position: Point) -> Self {
        Self::new(position, aperture.x, aperture.y, aperture.hole_diameter)
    }

    pub fn to_unit(&self, origin: Unit, to: Unit) -> Self {
        Self::new(
            self.position.to_unit(origin, to),
            convert_to_unit(self.x, origin, to),
            convert_to_unit(self.y, origin, to),
            self.hole_diameter.map(|d| convert_to_unit(d, origin, to)),
        )
    }
}

impl DrawableRaw for Obround {
    fn canvas_pos(&self) -> Point {
        self.position
    }

    fn draw_egui_plot(&self, ui: &mut egui_plot::PlotUi, colour: CanvasColour, selected: bool) {
        let mut points: Vec<[f64; 2]> = self.outline.iter().map(|p| [p.x(), p.y()]).collect();
        points.push(points[0]);

        let poly = Polygon::new(PlotPoints::from(points))
            .fill_color(colour.to_colour32(selected))
            .stroke(Stroke::NONE);
        ui.polygon(poly);
    }

    fn draw_egui(&self, ui: &mut Ui, selected: bool) {
        ui.painter().add(RectShape::filled(
            Rect::from_center_size(
                self.position.invert_y().into(),
                Vec2::new(self.rectangle.width as f32, self.rectangle.height as f32),
            ),
            Rounding::same(self.rounding as f32 / 2.),
            CanvasColour::Copper.to_colour32(selected),
        ));
    }

    fn to_paths(&self) -> ClipperPaths {
        ClipperPaths::new(vec![self.outline.clone()])
    }

    fn outline(&self) -> ClipperPath {
        self.outline.clone()
    }
}
