use eframe::{
    egui::{Rect, Rounding, Stroke, Ui, Vec2},
    epaint::RectShape,
};
use egui_plot::{PlotPoints, Polygon};

use crate::{
    application::CanvasColour,
    geometry::{ClipperPath, ClipperPaths},
};

use super::super::{
    point::{convert_to_unit, Point},
    DrawableRaw, Unit,
};

#[derive(Debug)]
pub struct Rectangle {
    position: Point,
    pub width: f64,
    pub height: f64,
    hole_diameter: Option<f64>,
    outline: ClipperPath,
}

impl Rectangle {
    pub fn from_aperture_rectangular(
        aperture: &gerber_types::Rectangular,
        position: Point,
    ) -> Self {
        Self {
            position,
            width: aperture.x,
            height: aperture.y,
            hole_diameter: aperture.hole_diameter,
            outline: vec![
                (position.x - aperture.x / 2., position.y - aperture.y / 2.),
                (position.x - aperture.x / 2., position.y + aperture.y / 2.),
                (position.x + aperture.x / 2., position.y + aperture.y / 2.),
                (position.x + aperture.x / 2., position.y - aperture.y / 2.),
                (position.x - aperture.x / 2., position.y - aperture.y / 2.),
            ]
            .into(),
        }
    }

    pub fn new(position: Point, width: f64, height: f64, hole_diameter: Option<f64>) -> Self {
        Self {
            position,
            width,
            height,
            hole_diameter,
            outline: vec![
                (position.x - width / 2., position.y - height / 2.),
                (position.x - width / 2., position.y + height / 2.),
                (position.x + width / 2., position.y + height / 2.),
                (position.x + width / 2., position.y - height / 2.),
            ]
            .into(),
        }
    }

    pub fn to_unit(&self, origin: Unit, to: Unit) -> Self {
        Self::new(
            self.position.to_unit(origin, to),
            convert_to_unit(self.width, origin, to),
            convert_to_unit(self.height, origin, to),
            self.hole_diameter.map(|d| convert_to_unit(d, origin, to)),
        )
    }
}

impl DrawableRaw for Rectangle {
    fn canvas_pos(&self) -> Point {
        self.position
            .shift_x(self.position.x / 2.)
            .shift_y(self.position.y / 2.)
    }

    fn draw_egui_plot(&self, ui: &mut egui_plot::PlotUi, colour: CanvasColour, selected: bool) {
        let mut points: Vec<[f64; 2]> = self.outline.iter().map(|p| [p.x(), p.y()]).collect();
        points.push(points[0]);

        let poly = Polygon::new(PlotPoints::from(points))
            .fill_color(colour.to_colour32(selected))
            .stroke(Stroke::NONE);
        ui.polygon(poly);
    }

    fn draw_egui(&self, ui: &mut Ui, selected: bool) {
        ui.painter().add(RectShape::filled(
            Rect::from_center_size(
                self.position.invert_y().into(),
                Vec2::new(self.width as f32, self.height as f32),
            ),
            Rounding::ZERO,
            CanvasColour::Copper.to_colour32(selected),
        ));
    }

    fn to_paths(&self) -> ClipperPaths {
        ClipperPaths::new(vec![self.outline.clone()])
    }

    fn outline(&self) -> ClipperPath {
        self.outline.clone()
    }
}
