use std::f64::consts::PI;

use super::point::Point;

const CIRCLE_SEGMENTS: u32 = 512;

pub fn semi_circle(center: Point, diameter: f64, tilt: f64) -> Vec<(f64, f64)> {
    (0..CIRCLE_SEGMENTS / 2)
        .step_by(1)
        .map(|i| {
            let angle = (i as f64 / (CIRCLE_SEGMENTS / 2) as f64) * PI + tilt * (PI / 180.);
            (
                angle.cos() * diameter / 2. + center.x,
                angle.sin() * diameter / 2. + center.y,
            )
        })
        .collect()
}

pub enum CircleSegment {
    North,
    East,
    South,
    West,
    Full,
}

pub fn create_circular_path(
    position: &Point,
    diameter: f64,
    segment: CircleSegment,
) -> Vec<(f64, f64)> {
    let segments: Vec<u32> = match segment {
        CircleSegment::North => (CIRCLE_SEGMENTS / 4..CIRCLE_SEGMENTS - CIRCLE_SEGMENTS / 4)
            .step_by(1)
            .collect(),
        CircleSegment::East => (0..CIRCLE_SEGMENTS / 2 + 1).step_by(1).collect(),
        CircleSegment::South => (CIRCLE_SEGMENTS - CIRCLE_SEGMENTS / 4..CIRCLE_SEGMENTS)
            .step_by(1)
            .chain((0..CIRCLE_SEGMENTS / 4).step_by(1))
            .collect(),
        CircleSegment::West => (CIRCLE_SEGMENTS / 2 - 1..CIRCLE_SEGMENTS)
            .step_by(1)
            .collect(),
        CircleSegment::Full => (0..CIRCLE_SEGMENTS).step_by(1).collect(),
    };
    segments
        .iter()
        .map(|&i| {
            let angle = (i as f64 / CIRCLE_SEGMENTS as f64) * 2.0 * PI;
            (
                angle.sin() * diameter / 2. + position.x,
                angle.cos() * diameter / 2. + position.y,
            )
        })
        .collect()
}

#[derive(Debug, PartialEq)]
pub enum Orientation {
    Clockwise,
    CounterClockwise,
    CoLinear,
}

// To find orientation of ordered triplet (p, q, r).
// https://www.geeksforgeeks.org/check-if-two-given-line-segments-intersect/
pub fn orientation(p: impl Into<Point>, q: impl Into<Point>, r: impl Into<Point>) -> Orientation {
    let (p, q, r) = (p.into(), q.into(), r.into());

    // See https://www.geeksforgeeks.org/orientation-3-ordered-points/
    // for details of below formula.
    let val = (q.y - p.y) * (r.x - q.x) - (q.x - p.x) * (r.y - q.y);

    match val {
        0. => Orientation::CoLinear,
        x if x > 0. => Orientation::Clockwise,
        _ => Orientation::CounterClockwise,
    }
}

// Given three collinear points p, q, r, the function checks if
// point q lies on line segment 'pr'
pub fn on_segment(p: impl Into<Point>, q: impl Into<Point>, r: impl Into<Point>) -> bool {
    let (p, q, r) = (p.into(), q.into(), r.into());

    q.x <= greater_val(p.x, r.x)
        && q.x >= lower_val(p.x, r.x)
        && q.y <= greater_val(p.y, r.y)
        && q.y >= lower_val(p.y, r.y)
}

fn greater_val(a: f64, b: f64) -> f64 {
    if a > b {
        a
    } else {
        b
    }
}

fn lower_val(a: f64, b: f64) -> f64 {
    if a < b {
        a
    } else {
        b
    }
}

// The main function that returns true if line segment 'p1q1'
// and 'p2q2' intersect.
// https://www.geeksforgeeks.org/check-if-two-given-line-segments-intersect/
pub fn do_intersect(
    p1: impl Into<Point>,
    q1: impl Into<Point>,
    p2: impl Into<Point>,
    q2: impl Into<Point>,
) -> bool {
    let (p1, q1, p2, q2) = (p1.into(), q1.into(), p2.into(), q2.into());

    // Find the four orientations needed for general and
    // special cases
    let o1 = orientation(p1, q1, p2);
    let o2 = orientation(p1, q1, q2);
    let o3 = orientation(p2, q2, p1);
    let o4 = orientation(p2, q2, q1);

    // General case
    if o1 != o2 && o3 != o4 {
        return true;
    }

    // Special Cases
    // p1, q1 and p2 are collinear and p2 lies on segment p1q1
    if o1 == Orientation::CoLinear && on_segment(p1, p2, q1) {
        return true;
    }

    // p1, q1 and q2 are collinear and q2 lies on segment p1q1
    if o2 == Orientation::CoLinear && on_segment(p1, q2, q1) {
        return true;
    };

    // p2, q2 and p1 are collinear and p1 lies on segment p2q2
    if o3 == Orientation::CoLinear && on_segment(p2, p1, q2) {
        return true;
    };

    // p2, q2 and q1 are collinear and q1 lies on segment p2q2
    if o4 == Orientation::CoLinear && on_segment(p2, q1, q2) {
        return true;
    };

    false // Doesn't fall in any of the above cases
}
