pub mod elements;
pub mod gerber;
mod helpers;
pub mod point;
mod union;

use std::fmt::{Debug, Display};

use clipper2::{Bounds, Path, Paths, PointScaler};
use eframe::egui::Ui;
use egui_plot::PlotUi;
use elements::{linepath::LinePath, Element};

use point::Point;

use crate::application::CanvasColour;

pub struct Geometry {
    pub outline_union: ClipperPaths,
    pub apertures: Vec<Element>,
    pub paths: Vec<LinePath>,
    pub units: Unit,
}

impl Geometry {
    pub fn to_unit(self, to: Unit) -> Self {
        Self {
            outline_union: self
                .outline_union
                .iter()
                .map(|p| {
                    p.iter()
                        .map(|p| (&Point::from(p).to_unit(self.units, to)).into())
                        .collect()
                })
                .collect(),
            apertures: self
                .apertures
                .iter()
                .map(|a| a.to_unit(self.units, to))
                .collect(),
            paths: self
                .paths
                .iter()
                .map(|l| l.to_unit(self.units, to))
                .collect(),
            units: to,
        }
    }
}

#[derive(Debug, PartialEq, Clone, Copy)]
pub enum Unit {
    Millimeters,
    Inches,
}

impl Default for Unit {
    fn default() -> Self {
        Self::Millimeters
    }
}

impl From<gerber_types::Unit> for Unit {
    fn from(value: gerber_types::Unit) -> Self {
        match value {
            gerber_types::Unit::Inches => Self::Inches,
            gerber_types::Unit::Millimeters => Self::Millimeters,
        }
    }
}

impl Display for Unit {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Unit::Millimeters => "mm",
                Unit::Inches => "in",
            }
        )
    }
}

/// Scale by 10000.
#[derive(Debug, Copy, Clone, PartialEq, Hash)]
pub struct Micro;

impl PointScaler for Micro {
    const MULTIPLIER: f64 = 10000.0;
}

pub type ClipperPoint = clipper2::Point<Micro>;
pub type ClipperPath = Path<Micro>;
pub type ClipperPaths = Paths<Micro>;
pub type ClipperBounds = Bounds<Micro>;

pub trait DrawableRaw {
    fn canvas_pos(&self) -> Point;
    fn draw_egui(&self, ui: &mut Ui, selected: bool);
    fn draw_egui_plot(&self, ui: &mut PlotUi, colour: CanvasColour, selected: bool);
    fn to_paths(&self) -> ClipperPaths;
    fn outline(&self) -> ClipperPath;
}

fn canvas_position_from_gerber(gerber_position: &Point, offset: Point) -> Point {
    gerber_position.shift_x(offset.x).shift_y(offset.y)
}
