use clipper2::{Bounds, EndType, JoinType, One, Paths};
use eframe::{
    egui::{Rect, Shape, Stroke},
    epaint::{PathShape, PathStroke},
};
use egui_plot::{PlotBounds, PlotItem, PlotPoint, Polygon};

use crate::{
    application::CanvasColour,
    geometry::{elements::circle::Circle, point::Point, ClipperBounds, ClipperPaths, Unit},
};

#[derive(Debug, Clone)]
pub enum GeometryType {
    Paths(ClipperPaths),
    Points(Vec<Point>),
}

#[derive(Debug)]
pub struct OutlineGeometry {
    // pub path: ClipperPaths,
    // pub points: Vec<Point>,
    items: GeometryType,
    pub stroke: f32,
    pub unit: Unit,
    pub bounds_from: String,
    pub bounding_box: ClipperBounds,
}

impl OutlineGeometry {
    pub fn new(
        outline: &ClipperPaths,
        stroke: f32,
        unit: Unit,
        bounds_from: &str,
        bounds: ClipperBounds,
    ) -> Self {
        // inflate given path
        let outline = outline
            .clone()
            .inflate((stroke / 2.).into(), JoinType::Round, EndType::Polygon, 2.0)
            .simplify(0.01, false);
        Self {
            // path: outline,
            // points: Vec::new(),
            items: GeometryType::Paths(outline),
            stroke,
            unit,
            bounds_from: bounds_from.into(),
            bounding_box: bounds,
        }
    }

    pub fn new_no_inflate(
        outline: &ClipperPaths,
        stroke: f32,
        unit: Unit,
        bounds_from: &str,
        bounds: ClipperBounds,
    ) -> Self {
        Self {
            // path: outline.clone(),
            // points: Vec::new(),
            items: GeometryType::Paths(outline.clone()),
            stroke,
            unit,
            bounds_from: bounds_from.into(),
            bounding_box: bounds,
        }
    }

    pub fn point_marker(
        points: Vec<Point>,
        stroke: f32,
        unit: Unit,
        bounds_from: &str,
        bounds: ClipperBounds,
    ) -> Self {
        Self {
            // path: Paths::new(vec![]),
            // points,
            items: GeometryType::Points(points),
            stroke,
            unit,
            bounds_from: bounds_from.into(),
            bounding_box: bounds,
        }
    }

    pub fn paths(&self) -> ClipperPaths {
        match &self.items {
            GeometryType::Paths(paths) => paths.clone(),
            GeometryType::Points(_) => ClipperPaths::new(vec![]),
        }
    }

    pub fn points(&self) -> Vec<Point> {
        match &self.items {
            GeometryType::Paths(_) => vec![],
            GeometryType::Points(p) => p.clone(),
        }
    }
}

pub struct OutlineShape {
    stroke: f32,
    selected: bool,
    colour: CanvasColour,
    items: GeometryType,
    bounds: ClipperBounds,
}

// impl OutlineShape {
//     pub fn new_from_geometry(
//         geometry: &OutlineGeometry,
//         colour: CanvasColour,
//         selected: bool,
//     ) -> Self {
//         Self {
//             stroke: geometry.stroke,
//             selected,
//             items: geometry.items.clone(),
//             bounds: geometry.bounding_box,
//             colour,
//         }
//     }
// }

// impl PlotItem for OutlineShape {
//     fn shapes(
//         &self,
//         _ui: &eframe::egui::Ui,
//         transform: &egui_plot::PlotTransform,
//         shapes: &mut Vec<eframe::egui::Shape>,
//     ) {
//         match &self.items {
//             GeometryType::Paths(paths) => {
//                 for path in paths.iter() {
//                     let values_tf: Vec<_> = path
//                         .iter()
//                         .map(|v| transform.position_from_point(&PlotPoint::from(Point::from(v))))
//                         .collect();

//                     let shape = PathShape::closed_line(
//                         values_tf.clone(),
//                         PathStroke::new(20., self.colour.to_colour32(self.selected)),
//                     );

//                     shapes.push(shape.into());
//                 }
//             }
//             GeometryType::Points(points) => {
//                 for point in points {
//                     // draw outline of hole
//                     let points = Circle::circle_segment_points(
//                         Point::from(transform.position_from_point(&PlotPoint::from(*point))),
//                         self.stroke.into(),
//                         1.,
//                         0.,
//                     )
//                     .into_iter()
//                     .map(|p| Point::from(p).into())
//                     .collect();

//                     let polygon = Shape::convex_polygon(
//                         points,
//                         self.colour.to_colour32(self.selected),
//                         Stroke::NONE,
//                     );
//                     shapes.push(polygon);
//                 }
//                 // TODO draw point circles
//             }
//         }

//         // todo!()
//     }

//     fn initialize(&mut self, x_range: std::ops::RangeInclusive<f64>) {
//         {}
//     }

//     fn name(&self) -> &str {
//         "Test"
//     }

//     fn color(&self) -> eframe::egui::Color32 {
//         self.colour.to_colour32(self.selected)
//     }

//     fn highlight(&mut self) {
//         {}
//     }

//     fn highlighted(&self) -> bool {
//         false
//     }

//     fn allow_hover(&self) -> bool {
//         false
//     }

//     fn geometry(&self) -> egui_plot::PlotGeometry<'_> {
//         todo!()
//     }

//     fn bounds(&self) -> PlotBounds {
//         PlotBounds::from_min_max(self.bounds.min.into(), self.bounds.max.into())
//     }

//     fn id(&self) -> Option<eframe::egui::Id> {
//         None
//     }
// }
